program Initial_Data
! 3D initial data solver for Brill wave / black hole spacetime
! Uses red/black Gauss-Seidel SOR
! David Rideout  June 1996
! to compile: f90 initdata.f90 -o initdata -cpp

! Currently 65^3 is the largest grid we can run on a kestrel (n=32)
! Could eliminate Psi and r2 to free up some memory usage...
! Maybe even red if it uses a significant amount of memory.
! Eliminating forall's allows xlf90 compiler ==> 256 Mb on merlins
!						(~50 Mb on kestrels)

#define DEBUG 0
#if DEBUG
#define D
#else
#define D !
#endif

implicit none
double precision, dimension(:,:,:), allocatable :: Psiold, Psinew, Psi, f, r2
logical, dimension(:,:,:), allocatable :: red !distinguishes two 'colors'
double precision :: test=1.0, error, tol, w, h
integer :: i, j, k, n, iter=1
logical :: theory=.false.

! Read parameters from input file
open(1, file='init.in')
read(1,*) n  ! (2n+1)^3 data points...
read(1,*) w, xmax, tol
close(1)
!print *, n, w, xmax, tol  ! to ensure parameters read properly
if (n<0) then
   theory=.true.
   n=-n
endif

! Allocate arrays
allocate(Psiold(-n:n,-n:n,-n:n), Psinew(-n:n,-n:n,-n:n), Psi(-n:n,-n:n,-n:n), &
  red(-n:n,-n:n,-n:n), f(-n:n,-n:n,-n:n), r2(-n:n,-n:n,-n:n))

! Generate red/black checkerboard:
red=.false.
forall(i=-n+1:n-1, j=-n+1:n-1, k=-n+1:n-1, mod(i+j+k,2)==0) red(i,j,k)=.true.

D do k=-n, n
D   print *, 'z index is ',k
D      do j=-n, n	!shows the red/black matrix (used for debugging)
D		print *, red(:,j,k)
D      end do
D   end do

! Generate theoretical solution and boundary values:
h=xmax/n
forall(i=-n:n, j=-n:n, k=-n:n) r2(i,j,k) = (i**2 + j**2 + k**2)*h**2
Psi=exp(-r2)+1
Psiold=Psi
do i = -n+1, n-1
do j = -n+1, n-1
do k = -n+1, n-1
   Psiold(i,j,k)  = 1 
end do
end do
end do

Psinew=Psiold !defines boundaries in Psinew

D print *,'Theoretical solution:'
D do k=-n, n
D   print *, 'z index is ',k
D      do j=-n, n
D	 print '(5F12.9)', Psi(:,j,k)
D      end do
D   end do
D print *,''

! Generate 'source term':
f = 2*exp(-r2)*(2*r2-3)
D print *,'source term f:'
D do k=-n, n
D   print *, 'z index is ',k
D      do j=-n, n
D	 print '(5F12.9)', f(:,j,k)
D      end do
D   end do
D print *,''


! Perform Gauss-Seidel iteration:
do while (test> tol)

D print *,'Psinew:'
D do k=-n, n
D   print *, 'z index is ',k
D      do j=-n, n
D	 print '(5F12.9)', Psinew(:,j,k)
D      end do
D   end do
D print *,''

   where (red) Psinew(-n+1:n-1,-n+1:n-1,-n+1:n-1) = &
     w*(Psiold(-n:n-2,-n+1:n-1,-n+1:n-1) + Psiold(-n+2:n,-n+1:n-1,-n+1:n-1) + &
     Psiold(-n+1:n-1,-n:n-2,-n+1:n-1) + Psiold(1:x-1,-n+2:n,-n+1:n-1) + &
     Psiold(-n+1:n-1,-n+1:n-1,-n:n-2) + Psiold(-n+1:n-1,-n+1:n-1,-n+2:n) &
     -h**2*f(-n+1:n-1,-n+1:n-1,-n+1:n-1))/6 + &
     (1-w)*Psiold(-n+1:n-1,-n+1:n-1,-n+1:n-1)
   where (.not. red) Psinew(-n+1:n-1,-n+1:n-1,-n+1:n-1) = &
     w*(Psinew(-n:n-2,-n+1:n-1,-n+1:n-1) + Psinew(-n+2:n,-n+1:n-1,-n+1:n-1) + &
     Psinew(-n+1:n-1,-n:n-2,-n+1:n-1) + Psinew(1:x-1,-n+2:n,-n+1:n-1) + &
     Psinew(-n+1:n-1,-n+1:n-1,-n:n-2) + Psinew(-n+1:n-1,-n+1:n-1,-n+2:n) &
     -h**2*f(-n+1:n-1,-n+1:n-1,-n+1:n-1))/6 + &
     (1-w)*Psinew(-n+1:n-1,-n+1:n-1,-n+1:n-1)

   test=maxval(abs(Psinew(-n+1:n-1,-n+1:n-1,-n+1:n-1) - &
     Psiold(-n+1:n-1,-n+1:n-1,-n+1:n-1))) !max abs. dev.
   Psiold=Psinew !store result in Psiold for comparison with next iteration
   iter = iter + 1 !count number of iterations

	print *, test, 'test'

   if (test>1000) stop  ! Let's not sit here forever if it is going to diverge!
end do

! Output final results:
D print *,'Final result for Psi:'
D do k=-n, n
D   print *, 'z index is ',k
D   do j=-n, n
D      print '(5F12.9)', Psinew(:,j,k)
D   end do
D end do

error=sum(abs(Psinew-Psi))/(2*n+1)**3 !average deviation from theoretical solution
print *, 'Solution took',iter,' iterations.'
print *, 'Average error from theoretical solution is', error,' .'

! Output for convergence test
do i=-n, n
   if (theory) then
      print '(F12.9, F19.15)', i*h, Psi(i,0,0)
   else
      print '(F12.9, F19.15)', i*h, Psinew(i,0,0)
   endif
end do

#undef D
end program Initial_Data




