! 3D initial data solver for Brill wave / black hole spacetime
! Uses red/black Gauss-Seidel SOR
! David Rideout  July 1996
! to compile: f90 initdata.f90 -o initdata -cpp

! Currently 129^3 is the largest grid we can run on a kestrel (nx=64)
! Can (and will) eliminate tPsi to free up some memory usage
! Eliminating forall's allows xlf90 compiler ==> 256 Mb on merlins
!						(~50 Mb on kestrels)




module data
implicit none
integer :: nx, ny, nz, i, j, k, l
real*8 , dimension(:,:,:), allocatable, target :: Psi, res, f
 real*8 , dimension(:,:,:), allocatable :: tPsi	! for quick convergence tests; delete later ***
real*8 :: h, xmax
real*8, dimension(:,:,:), pointer :: m
end module data
!------------------------------------------------------------------------
program Initial_Data
use data
implicit none
real*8 :: test=1.0d0, error, normres=10.0d0, normresb, tol, w
integer :: iter=0
logical :: theory=.false.

! Read parameters from input file
call parameters(w, tol)
h=xmax/nx
 print *, nx, ny, nz, w, xmax, tol, h  ! to ensure parameters read properly
if (nx<0) then  ! (used for convergence testing)
   theory=.true.
   nx=-nx
endif

! Allocate arrays
allocate(Psi(-nx:nx,-ny:ny,-nz:nz), res(-nx:nx,-ny:ny,-nz:nz), &
  f(-nx:nx,-ny:ny,-nz:nz))

! Generate theoretical solution & source term
call init_data

! Perform Gauss-Seidel iteration:
 print *, 'Begining iteration...'
res=0	! set boundary residuals to zero
do while (normres> tol)
 print *,'red residuals...'	! lets believe that all the interior stuff
   call residual(.true.)	! works for now...
 m=> res
! print *,'relaxing...'
!   call relax(w, 'r')
 print *,'Psi:'
 m=> Psi
 print *,'black residuals...'
   call residual(.false.)
 m=> res
!   call relax(w, 'b')

 print *,'residuals at the outer boundary...'

 print *,'Psi used to calculate residual at outer boundary:'
 m => Psi

!Psi=1; res=1; f=0; Psi(0,0,0)=0
   call res_ob
 print *,'residual:'
m=> res
 call dump(-nx,-ny,'zero.dat')
 call dump(nx,ny,'one.dat')
stop

 print *,'final relaxing for sweep'
   call relax(w, 'o')

 print *,'Psi:'
m=>Psi

   m => res
   call l2n(normres)
   print *, 'overall norm of residual:', normres

   call l2n_bdry(normresb)
   print *, 'norm of residual on outer boundary:', normresb

 print *,'Deviation of edge value from theory:', Psi(-nx,0,0)-tPsi(-nx,0,0)
 print *,'tPsi(-nx,0,0) = ', tPsi(-nx,0,0)
!	stop

   iter = iter + 1 !count number of iterations
   if (mod(iter, 100)==1) print *, 'iter',iter,'residue norm',normres
   if (normres>100000) stop
                     ! Let's not sit here forever if it is going to diverge!
end do

! Output final results:
 print *,'Final result for Psi:'
 m=>Psi

 error=sum(abs(Psi-tPsi))/(2*nx+1)/(2*ny+1)/(2*nz+1)
                                !average deviation from theoretical solution
print *, 'x grid size:', 2*nx+1
print *, 'Solution took',iter,' iterations.'
print *, 'Norm of Residual is',normres,' .'
 print *, 'Average error from theoretical solution is', error,' .'

! Output for convergence test
!do i=-n, n
!   if (theory) then
!      print '(F12.9, F19.15)', i*h, tPsi(i,0,0)
!   else
!      print '(F12.9, F19.15)', i*h, Psi(i,0,0)
!   endif
!end do

end program Initial_Data
!------------------------------------------------------------------------
subroutine parameters(w, tol)
use data
implicit none
real*8 :: w, tol
character(10) :: check="" ! checks for errors

open(1, file='init.in')
read(1,*) nx, ny, nz  ! (2nx+1)(2ny+1)(2nz+1) data points...
read(1,*) check
if (.not. check=="hithere") then
   print *,"Error in init.in ; program terminating"
   stop
end if
read(1,*) w, xmax, tol
close(1)

end subroutine parameters
!------------------------------------------------------------------------
subroutine init_data
use data
implicit none

! Generate theoretical solution and boundary values:
do i=-nx, nx
   do j=-ny, ny
      do k=-nz, nz
         res(i,j,k) = (i**2 + j**2 + k**2)*h**2
      end do     ! res is radius squared for now
   end do
end do

Psi=exp(-res)+1. ! Psi holds the theoretical solution for now

 allocate(tPsi(-nx:nx,-ny:ny,-nz:nz))
                           !  for quick convergence tests; delete later ***
 tPsi=Psi	

! print *,'Theoretical solution:'
! m=>Psi ! dump out Psi, m is a pointer used in dump
! call dump
! r squared and Psi seem to work fine 8/8/96

!Psi=0
!Psi(-nx+1:nx-1,-ny+1:ny-1,-nz+1:nz-1)=1
!Psi(-nx+2:nx-2,-ny+2:ny-2,-nz+2:nz-2)=2   ! Initial guess for Psi

! Generate 'source term':
f = 2.*exp(-res)*(2.*res-3.)

! Generate source term on boundaries:
do i=-nx, nx
   do j=-ny, ny
      do k=-nz, nz
         if (.not. (abs(i)<nx .and. abs(j)<ny .and. abs(k)<nz)) &
              f(i,j,k)=exp(-res(i,j,k))*((1/sqrt(res(i,j,k))) &
              - 2*sqrt(res(i,j,k)))
      end do     ! res is radius squared for now
   end do
end do

 print *,'initial guess for Psi:'
 m=> Psi
 print *,'source term f/g:'
 m=> f
! Source term is correctly generated 8/8/96

end subroutine init_data
!------------------------------------------------------------------------
subroutine residual(red)
use data
implicit none
logical :: red

if (red) then
   do i=-nx+1, nx-1
      do j=-ny+1, ny-1
         do k=-nz+1, nz-1
            if (mod(i+j+k,2)==0) res(i,j,k) = &
                 (Psi(i-1,  j,k  ) + Psi(i+1,j  ,k  ) + &
                  Psi(i  ,j-1,k  ) + Psi(i  ,j+1,k  ) + &
                  Psi(i  ,j  ,k-1) + Psi(i  ,j  ,k+1) - &
                  6*Psi(i,j,k))/h/h - f(i,j,k)
         end do
      end do
   end do
else
   do i=-nx+1, nx-1
      do j=-ny+1, ny-1
         do k=-nz+1, nz-1
            if (.not. mod(i+j+k,2)==0) res(i,j,k) = &
                 (Psi(i-1,  j,k  ) + Psi(i+1,j  ,k  ) + &
                  Psi(i  ,j-1,k  ) + Psi(i  ,j+1,k  ) + &
                  Psi(i  ,j  ,k-1) + Psi(i  ,j  ,k+1) - &
                  6*Psi(i,j,k))/h/h - f(i,j,k)
         end do
      end do
   end do
endif

end subroutine residual
!------------------------------------------------------------------------------
subroutine res_ob
use data
implicit none
real*8 , dimension(:,:), allocatable :: r
real*8 , dimension(:), allocatable :: r1
real*8 :: rc

!if (red) then
!Forget red/black for now on the outer boundary...
!It will look just like the residual routine, generating an enormous amount of source code!!


!Psi(-nx,-ny+1:ny-1,-nz+1,nz-1): 			  -------x faces
 print *,'x-faces...'
allocate(r(-ny+1:ny-1,-nz+1:nz-1))
do j=-ny+1, ny-1
   do k=-nz+1, nz-1
      r(j,k) = h*sqrt(nx**2 + 1.0d0*j*j + 1.0d0*k*k)
   end do
end do
!r(-ny+1:ny-1, -nz+1:nz-1) = h*sqrt((nx)**2 + &
!  spread([-ny+1:ny-1]**2,2,2*nz-1) + spread([-nz+1:nz-1]**2,1,2*ny-1))
do j=-ny+1, ny-1
   do k=-nz+1, nz-1
      res(-nx,j,k) = &
           (-nx*(-25*Psi(-nx,j,k) +48*Psi(-nx+1,j,k) -36*Psi(-nx+2,j,k) &
           + 16*Psi(-nx+3,j,k) -3*Psi(-nx+4,j,k))/12                    &
           +j*(Psi(-nx,j+1,k)-Psi(-nx,j-1,k))/2                         &
           +k*(Psi(-nx,j,k+1)-Psi(-nx,j,k-1))/2   &   ! dPsi/dr             
           - 1 + Psi(-nx,j,k))/r(j,k) - f(-nx,j,k)

!Psi( nx,:,:):
      res(nx,j,k) = &
           (nx*(25*Psi(nx,j,k) - 48*Psi(nx-1,j,k) + 36*Psi(nx-2,j,k)   &
           - 16*Psi(nx-3,j,k) + 3*Psi(nx-4,j,k))/12                       &
           +j*(Psi(nx,j+1,k)-Psi(nx,j-1,k))/2                          &
           +k*(Psi(nx,j,k+1)-Psi(nx,j,k-1))/2   &   ! dPsi/dr             
           - 1 + Psi(nx,j,k))/r(j,k) - f(nx,j,k)
   end do
end do
deallocate(r)


!Psi(:,-ny,:):						--------y faces
 print *,'y faces...'
allocate(r(-nx+1:nx-1,-nz+1:nz-1))
do i=-nx+1, nx-1
   do k=-nz+1, nz-1
      r(i,k) = h*sqrt(i**2 + ny**2 + 1.0d0*k**2)
   end do
end do
!forall(i=-nx+1:nx-1, k=-nz+1:nz-1)
do i=-nx+1, nx-1
 do k=-nz+1, nz-1
  res(i,-ny,k) = &
  (i*(Psi(i+1,-ny,k)-Psi(i-1,-ny,k))/2 				&
  -ny*(-25*Psi(i,-ny,k) +48*Psi(i,-ny+1,k) -36*Psi(i,-ny+2,k)  &
    + 16*Psi(i,-ny+3,k) -3*Psi(i,-ny+4,k))/12                     &
  + k*(Psi(i,-ny,k+1)-Psi(i,-ny,k-1))/2   &   ! dPsi/dr             
  - 1 + Psi(i,-ny,k))/r(i,k) - f(i,-ny,k)

!Psi(:, ny,:)
!forall(i=-nx+1:nx-1, k=-nz+1:nz-1)
  res(i,ny,k) = &
  (i*(Psi(i+1,ny,k)-Psi(i-1,ny,k))/2                          &
  + ny*(25*Psi(i,ny,k) - 48*Psi(i,ny-1,k) + 36*Psi(i,ny-2,k)   &
    - 16*Psi(i,ny-3,k) + 3*Psi(i,ny-4,k))/12                     &
  +k*(Psi(i,ny,k+1)-Psi(i,ny,k-1))/2   &   ! dPsi/dr             
  - 1 + Psi(i,ny,k))/r(i,k) - f(i,ny,k)
 end do
end do
deallocate(r)


!Psi(:,:,-nz)						--------z faces
 print *,'z faces...'
allocate(r(-nx+1:nx-1,-ny+1:ny-1))
do i=-nx+1, nx-1
   do j=-ny+1, ny-1
      r(i,j) = h*sqrt(i**2 + 1.*j*j + nz**2)
   end do
end do
!forall(i=-nx+1:nx-1, j=-ny+1:ny-1)
do i=-nx+1, nx-1
 do j=-ny+1, ny-1
  res(i,j,-nz) = &
  (i*(Psi(i+1,j,-nz)-Psi(i-1,j,-nz))/2				&
  + j*(Psi(i,j+1,nz)-Psi(i,j-1,nz))/2				&
  - nz*(-25*Psi(i,j,-nz) +48*Psi(i,j,-nz+1) -36*Psi(i,j,-nz+2)  &
    + 16*Psi(i,j,-nz+3) -3*Psi(i,j,-nz+4))/12                     &
  - 1 + Psi(i,j,-nz))/r(i,j) - f(i,j,-nz)

!Psi(:,:, nz)
!forall(i=-nx+1:nx-1, j=-ny+1:ny-1)
  res(i,j,nz) = &
  (i*(Psi(i+1,j,nz)-Psi(i-1,j,nz))/2         &
  + j*(Psi(i,j+1,nz)-Psi(i,j-1,nz))/2   &
  + nz*(25*Psi(i,j,nz) - 48*Psi(i,j,nz-1) + 36*Psi(i,j,nz-2)   &
    - 16*Psi(i,j,nz-3) + 3*Psi(i,j,nz-4))/12                     &
  - 1 + Psi(i,j,nz))/r(i,j) - f(i,j,nz)
 end do
end do
deallocate(r)

 print *,'x edges...'
allocate(r1(-nx+1:nx-1))				!--------4 x edges
r1=h*sqrt((/(l,l=-nx+1,nx-1)/)**2 + 1.0d0*ny**2 + nz**2)
!forall(i=-nx+1:nx-1)
do i=-nx+1, nx-1
 print *,'first x edge'
  res(i,-ny,-nz) = &
  (i*(Psi(i+1,-ny,-nz)-Psi(i-1,-ny,-nz))			&
  -ny*(-25*Psi(i,-ny,-nz) + 48*Psi(i,-ny+1,-nz) - 36*Psi(i,-ny+2,-nz)  &
    + 16*Psi(i,-ny+3,-nz) - 3*Psi(i,-ny+4,-nz))/12    		&
  -nz*(-25*Psi(i,-ny,-nz) +48*Psi(i,-ny,-nz+1) -36*Psi(i,-ny,-nz+2)  &
    + 16*Psi(i,-ny,-nz+3) -3*Psi(i,-ny,-nz+4))/12			&
  - 1 + Psi(i,-ny,-nz))/r1(i) - f(i,-ny,-nz)

 print *,'second x edge'
!forall(i=-nx+1:nx-1)
  res(i,ny,-nz) = &
  (i*(Psi(i+1,ny,-nz)-Psi(i-1,ny,-nz))				&
  + ny*(25*Psi(i,ny,-nz) - 48*Psi(i,ny-1,-nz) + 36*Psi(i,ny-2,-nz)  &
    - 16*Psi(i,ny-3,-nz) + 3*Psi(i,ny-4,-nz))/12    		&
  - nz*(-25*Psi(i,ny,-nz) +48*Psi(i,ny,-nz+1) -36*Psi(i,ny,-nz+2)  &
    + 16*Psi(i,ny,-nz+3) -3*Psi(i,ny,-nz+4))/12			&
  - 1 + Psi(i,ny,-nz))/r1(i) - f(i,ny,-nz)

 print *,'third x edge'
!forall(i=-nx+1:nx-1)
  res(i,-ny,nz) = &
  (i*(Psi(i+1,-ny,nz)-Psi(i-1,-ny,nz))				&
  - ny*(-25*Psi(i,-ny,nz) + 48*Psi(i,-ny+1,nz) - 36*Psi(i,-ny+2,nz)  &
    + 16*Psi(i,-ny+3,nz) - 3*Psi(i,-ny+4,nz))/12    		&
  + nz*(25*Psi(i,-ny,nz) - 48*Psi(i,-ny,nz-1) + 36*Psi(i,-ny,nz-2)  &
    - 16*Psi(i,-ny,nz-3) + 3*Psi(i,-ny,nz-4))/12			&
  - 1 + Psi(i,-ny,nz))/r1(i) - f(i,-ny,nz)

 print *,'fourth x edge'
!forall(i=-nx+1:nx-1)
  res(i,ny,nz) = &
  (i*(Psi(i+1,ny,nz)-Psi(i-1,ny,nz))				&
  + ny*(25*Psi(i,ny,nz) - 48*Psi(i,ny-1,nz) + 36*Psi(i,ny-2,nz)  &
    - 16*Psi(i,ny-3,nz) + 3*Psi(i,ny-4,nz))/12			&
  + nz*(25*Psi(i,ny,nz) - 48*Psi(i,ny,nz-1) + 36*Psi(i,ny,nz-2)  &
    - 16*Psi(i,ny,nz-3) + 3*Psi(i,ny,nz-4))/12			&
  - 1 + Psi(i,ny,nz))/r1(i) - f(i,ny,nz)
end do
deallocate(r1)


 print *,'y edges...'
allocate(r1(-ny+1:ny-1))				!--------4 y edges
r1 = h*sqrt(1.0d0*nx**2 + (/(l,l=-ny+1,ny-1)/)**2 + nz**2)
!forall(j=-ny+1:ny-1)
do j=-ny+1, ny-1
  res(-nx,j,-nz) = &
  (-nx*(-25*Psi(-nx,j,-nz) + 48*Psi(-nx+1,j,-nz) - 36*Psi(-nx+2,j,-nz)  &
    + 16*Psi(-nx+3,j,-nz) - 3*Psi(-nx+4,j,-nz))/12    		&
  + j*(Psi(-nx,j+1,-nz) - Psi(-nx,j-1,-nz))			&
  -nz*(-25*Psi(-nx,j,-nz) + 48*Psi(-nx,j,-nz+1) - 36*Psi(-nx,j,-nz+2)  &
    + 16*Psi(-nx,j,-nz+3) -3*Psi(-nx,j,-nz+4))/12			&
  - 1 + Psi(-nx,j,-nz))/r1(j) - f(-nx,j,-nz)

!forall(j=-ny+1:ny-1)
  res(nx,j,-nz) = &
  (nx*(25*Psi(nx,j,-nz) - 48*Psi(nx-1,j,-nz) + 36*Psi(nx-2,j,-nz)  &
    - 16*Psi(nx-3,j,-nz) + 3*Psi(nx-4,j,-nz))/12    		&
  + j*(Psi(nx,j+1,-nz) - Psi(nx,j-1,-nz))				&
  -nz*(-25*Psi(nx,j,-nz) +48*Psi(nx,j,-nz+1) -36*Psi(nx,j,-nz+2)  &
    + 16*Psi(nx,j,-nz+3) -3*Psi(nx,j,-nz+4))/12			&
  - 1 + Psi(nx,j,-nz))/r1(j) - f(nx,j,-nz)

!forall(j=-ny+1:ny-1)
  res(-nx,j,nz) = &
  (-nx*(-25*Psi(-nx,j,nz) + 48*Psi(-nx+1,j,nz) - 36*Psi(-nx+2,j,nz)  &
    + 16*Psi(-nx+3,j,nz) - 3*Psi(-nx+4,j,nz))/12    		&
  + j*(Psi(-nx,j+1,nz) - Psi(-nx,j-1,nz))				&
  + nz*(25*Psi(-nx,j,nz) - 48*Psi(-nx,j,nz-1) + 36*Psi(-nx,j,nz-2)  &
    - 16*Psi(-nx,j,nz-3) + 3*Psi(-nx,j,nz-4))/12			&
  - 1 + Psi(-nx,j,nz))/r1(j) - f(-nx,j,nz)

!forall(j=-ny+1:ny-1)
  res(nx,j,nz) = &
  (nx*(25*Psi(nx,j,nz) - 48*Psi(nx-1,j,nz) + 36*Psi(nx-2,j,nz)  &
    - 16*Psi(nx-3,j,nz) + 3*Psi(nx-4,j,nz))/12			&
  + j*(Psi(nx,j+1,nz) - Psi(nx,j-1,nz))				&
  + nz*(25*Psi(nx,j,nz) - 48*Psi(nx,j,nz-1) + 36*Psi(nx,j,nz-2)  &
    - 16*Psi(nx,j,nz-3) + 3*Psi(nx,j,nz-4))/12			&
  - 1 + Psi(nx,j,nz))/r1(j) - f(nx,j,nz)
end do
deallocate(r1)


 print *,'z edges...'
allocate(r1(-nz+1:nz-1))				!--------4 z edges
r1 = h*sqrt(nx**2 + 1.0d0*ny**2 + (/(l,l=-nz+1,nz-1)/)**2)
!forall(k=-nz+1:nz-1)
do k=-nz+1, nz-1
  res(-nx,-ny,k) = &
  (-nx*(-25*Psi(-nx,-ny,k) + 48*Psi(-nx+1,-ny,k) - 36*Psi(-nx+2,-ny,k)  &
    + 16*Psi(-nx+3,-ny,k) - 3*Psi(-nx+4,-ny,k))/12    		&
  -ny*(-25*Psi(-nx,-ny,k) + 48*Psi(-nx,-ny+1,k) - 36*Psi(-nx,-ny+2,k)  &
    + 16*Psi(-nx,-ny+3,k) -3*Psi(-nx,-ny+4,-nz))/12			&
  + k*(Psi(-nx,-ny,k+1) - Psi(-nx,-ny,k-1))			&
  - 1 + Psi(-nx,-ny,k))/r1(k) - f(-nx,-ny,k)

!forall(k=-nz+1:nz-1)
  res(nx,-ny,k) = &
  (nx*(25*Psi(nx,-ny,k) - 48*Psi(nx-1,-ny,k) + 36*Psi(nx-2,-ny,k)  &
    - 16*Psi(nx-3,-ny,k) + 3*Psi(nx-4,-ny,k))/12    		&
  - ny*(-25*Psi(nx,-ny,k) +48*Psi(nx,-ny+1,k) -36*Psi(nx,-ny+2,k)  &
    + 16*Psi(nx,-ny+3,k) -3*Psi(nx,-ny+4,k))/12			&
  - k*(Psi(nx,-ny,k+1) - Psi(nx,-ny,k-1))				&
  - 1 + Psi(nx,-ny,k))/r1(k) - f(nx,-ny,k)

!forall(k=-nz+1:nz-1)
  res(-nx,ny,k) = &
  (-nx*(-25*Psi(-nx,ny,k) + 48*Psi(-nx+1,ny,k) - 36*Psi(-nx+2,ny,k)  &
    + 16*Psi(-nx+3,ny,k) - 3*Psi(-nx+4,ny,k))/12    		&
  + ny*(25*Psi(-nx,ny,k) - 48*Psi(-nx,ny-1,k) + 36*Psi(-nx,ny-2,k)  &
    - 16*Psi(-nx,ny-3,k) + 3*Psi(-nx,ny-4,k))/12			&
  + k*(Psi(-nx,ny,k+1) - Psi(-nx,ny,k-1))				&
  - 1 + Psi(-nx,ny,k))/r1(k) - f(-nx,ny,k)

!forall(k=-nz+1:nz-1)
  res(nx,ny,k) = &
  (nx*(25*Psi(nx,ny,k) - 48*Psi(nx-1,ny,k) + 36*Psi(nx-2,ny,k)  &
    - 16*Psi(nx-3,ny,k) + 3*Psi(nx-4,ny,k))/12			&
  + ny*(25*Psi(nx,ny,k) - 48*Psi(nx,ny-1,k) + 36*Psi(nx,ny-2,k)  &
    - 16*Psi(nx,ny-3,k) + 3*Psi(nx,ny-4,k))/12			&
  + k*(Psi(nx,ny,k+1) - Psi(nx,ny,k-1))				&
  - 1 + Psi(nx,ny,k))/r1(k) - f(nx,ny,k)
end do
deallocate(r1)


 print *,'z neg. corners...'
!						--------4 corners, z neg.
rc=h*sqrt(nx**2 + 1.0d0*ny**2 + nz**2)
res(-nx,-ny,-nz) = &
  (-nx*(-25*Psi(-nx,-ny,-nz) + 48*Psi(-nx+1,-ny,-nz) - 36*Psi(-nx+2,-ny,-nz) &
    + 16*Psi(-nx+3,-ny,-nz) - 3*Psi(-nx+4,-ny,-nz))/12 &
  -ny*(-25*Psi(-nx,-ny,-nz) + 48*Psi(-nx,-ny+1,-nz) - 36*Psi(-nx,-ny+2,-nz) &
    + 16*Psi(-nx,-ny+3,-nz) - 3*Psi(-nx,-ny+4,-nz))/12 &
  -nz*(-25*Psi(-nx,-ny,-nz) + 48*Psi(-nx,-ny,-nz+1) - 36*Psi(-nx,-ny,-nz+2) &
    + 16*Psi(-nx,-ny,-nz+3) - 3*Psi(-nx,-ny,-nz+4))/12 &
  - 1 + Psi(-nx,-ny,-nz))/rc - f(-nx,-ny,-nz)

res(nx,-ny,-nz) = &
  (nx*(25*Psi(nx,-ny,-nz) - 48*Psi(nx-1,-ny,-nz) + 36*Psi(nx-2,-ny,-nz) &
    - 16*Psi(nx-3,-ny,-nz) + 3*Psi(nx-4,-ny,-nz))/12 &
  -ny*(-25*Psi(nx,-ny,-nz) + 48*Psi(nx,-ny+1,-nz) - 36*Psi(nx,-ny+2,-nz) &
    + 16*Psi(nx,-ny+3,-nz) - 3*Psi(nx,-ny+4,-nz))/12 &
  -nz*(-25*Psi(nx,-ny,-nz) + 48*Psi(nx,-ny,-nz+1) - 36*Psi(nx,-ny,-nz+2) &
    + 16*Psi(nx,-ny,-nz+3) - 3*Psi(nx,-ny,-nz+4))/12 &
  - 1 + Psi(nx,-ny,-nz))/rc - f(nx,-ny,-nz)

res(-nx,ny,-nz) = &
  (-nx*(-25*Psi(-nx,ny,-nz) + 48*Psi(-nx+1,ny,-nz) - 36*Psi(-nx+2,ny,-nz) &
    + 16*Psi(-nx+3,ny,-nz) - 3*Psi(-nx+4,ny,-nz))/12 &
  + ny*(25*Psi(-nx,ny,-nz) - 48*Psi(-nx,ny-1,-nz) + 36*Psi(-nx,ny-2,-nz) &
    - 16*Psi(-nx,ny-3,-nz) + 3*Psi(-nx,ny-4,-nz))/12 &
  - nz*(-25*Psi(-nx,ny,-nz) + 48*Psi(-nx,ny,-nz+1) - 36*Psi(-nx,ny,-nz+2) &
    + 16*Psi(-nx,ny,-nz+3) - 3*Psi(-nx,ny,-nz+4))/12 &
  - 1 + Psi(-nx,ny,-nz))/rc - f(-nx,ny,-nz)

res(nx,ny,-nz) = &
  (nx*(25*Psi(nx,ny,-nz) - 48*Psi(nx-1,ny,-nz) + 36*Psi(nx-2,ny,-nz) &
    - 16*Psi(nx-3,ny,-nz) + 3*Psi(nx-4,ny,-nz))/12 &
  + ny*(25*Psi(nx,ny,-nz) - 48*Psi(nx,ny-1,-nz) + 36*Psi(nx,ny-2,-nz) &
    - 16*Psi(nx,ny-3,-nz) + 3*Psi(nx,ny-4,-nz))/12 &
  - nz*(-25*Psi(nx,ny,-nz) + 48*Psi(nx,ny,-nz+1) - 36*Psi(nx,ny,-nz+2) &
    + 16*Psi(nx,ny,-nz+3) - 3*Psi(nx,ny,-nz+4))/12 &
  - 1 + Psi(nx,ny,-nz))/rc - f(nx,ny,-nz)


 print *,'z pos. corners...'
!						--------4 corners, z pos.
res(-nx,-ny,nz) = &
  (-nx*(-25*Psi(-nx,-ny,nz) + 48*Psi(-nx+1,-ny,nz) - 36*Psi(-nx+2,-ny,nz) &
    + 16*Psi(-nx+3,-ny,nz) - 3*Psi(-nx+4,-ny,nz))/12 &
  - ny*(-25*Psi(-nx,-ny,nz) + 48*Psi(-nx,-ny+1,nz) - 36*Psi(-nx,-ny+2,nz) &
    + 16*Psi(-nx,-ny+3,nz) - 3*Psi(-nx,-ny+4,nz))/12 &
  + nz*(25*Psi(-nx,-ny,nz) - 48*Psi(-nx,-ny,nz-1) + 36*Psi(-nx,-ny,nz-2) &
    - 16*Psi(-nx,-ny,nz-3) + 3*Psi(-nx,-ny,nz-4))/12 &
  - 1 + Psi(-nx,-ny,nz))/rc - f(-nx,-ny,nz)

res(nx,-ny,nz) = &
  (nx*(25*Psi(nx,-ny,nz) - 48*Psi(nx-1,-ny,nz) + 36*Psi(nx-2,-ny,nz) &
    - 16*Psi(nx-3,-ny,nz) + 3*Psi(nx-4,-ny,nz))/12 &
  - ny*(-25*Psi(nx,-ny,nz) + 48*Psi(nx,-ny+1,nz) - 36*Psi(nx,-ny+2,nz) &
    + 16*Psi(nx,-ny+3,nz) - 3*Psi(nx,-ny+4,nz))/12 &
  + nz*(25*Psi(nx,-ny,nz) - 48*Psi(nx,-ny,nz-1) + 36*Psi(nx,-ny,nz-2) &
    - 16*Psi(nx,-ny,nz-3) + 3*Psi(nx,-ny,nz-4))/12 &
  - 1 + Psi(nx,-ny,nz))/rc - f(nx,-ny,nz)

res(-nx,ny,nz) = &
  (-nx*(-25*Psi(-nx,ny,nz) + 48*Psi(-nx+1,ny,nz) - 36*Psi(-nx+2,ny,nz) &
    + 16*Psi(-nx+3,ny,nz) - 3*Psi(-nx+4,ny,nz))/12 &
  + ny*(25*Psi(-nx,ny,nz) - 48*Psi(-nx,ny-1,nz) + 36*Psi(-nx,ny-2,nz) &
    - 16*Psi(-nx,ny-3,nz) + 3*Psi(-nx,ny-4,nz))/12 &
  + nz*(25*Psi(-nx,ny,nz) - 48*Psi(-nx,ny,nz-1) + 36*Psi(-nx,ny,nz-2) &
    - 16*Psi(-nx,ny,nz-3) + 3*Psi(-nx,ny,nz-4))/12 &
  - 1 + Psi(-nx,ny,nz))/rc - f(-nx,ny,nz)

res(nx,ny,nz) = &
  (nx*(25*Psi(nx,ny,nz) - 48*Psi(nx-1,ny,nz) + 36*Psi(nx-2,ny,nz) &
    - 16*Psi(nx-3,ny,nz) + 3*Psi(nx-4,ny,nz))/12 &
  + ny*(25*Psi(nx,ny,nz) - 48*Psi(nx,ny-1,nz) + 36*Psi(nx,ny-2,nz) &
    - 16*Psi(nx,ny-3,nz) + 3*Psi(nx,ny-4,nz))/12 &
  + nz*(25*Psi(nx,ny,nz) - 48*Psi(nx,ny,nz-1) + 36*Psi(nx,ny,nz-2) &
    - 16*Psi(nx,ny,nz-3) + 3*Psi(nx,ny,nz-4))/12 &
  - 1 + Psi(nx,ny,nz))/rc - f(nx,ny,nz)

end subroutine res_ob
!------------------------------------------------------------------------------
subroutine relax(w, red)
use data
implicit none
real*8 :: w
character :: red

select case(red)
 case('r')
 print *,'red relaxing'
   do i=-nx+1, nx-1
      do j=-ny+1, ny-1
         do k=-nz+1, nz-1
            if (mod(i+j+k,2)==0) Psi(i,j,k) = w*h*h*res(i,j,k)/6 + Psi(i,j,k)
         end do
      end do
   end do
 case('b')
 print *,'black relaxing'
   do i=-nx+1, nx-1
      do j=-ny+1, ny-1
         do k=-nz+1, nz-1
            if (.not. mod(i+j+k,2)==0) Psi(i,j,k) = &
                 w*h*h*res(i,j,k)/6 + Psi(i,j,k)
         end do
      end do
   end do
 case('o')
 print *,'out bd. relaxing'
 do i=-nx, nx
    do j=-ny, ny
       do k=-nz, nz
          if (.not. (abs(i)<nx .and. abs(j)<ny .and. abs(k)<nz)) &
               Psi(i,j,k) = w*h*h*res(i,j,k)/6 + Psi(i,j,k)               
       end do
    end do
 end do

! I believe that this is wrong below
!   Psi(-nx,-ny:ny,-nz:nz) = w*h*h*res(-nx,-ny:ny,-nz:nz)/6 + &
!        Psi(-nx,-ny:ny,-nz:nz)
!   Psi(nx,-ny:ny,-nz:nz) = w*h*h*res(nx,-ny:ny,-nz:nz)/6 + &
!        Psi(nx,-ny:ny,-nz:nz)
!   Psi(-nx:nx,-ny,-nz:nz) = w*h*h*res(-nx:nx,-ny,-nz:nz)/6 + &
!        Psi(-nx:nx,-ny,-nz:nz)
!   Psi(-nx:nx,ny,-nz:nz) = w*h*h*res(-nx:nx,ny,-nz:nz)/6 + &
!        Psi(-nx:nx,ny,-nz:nz)
!   Psi(-nx:nx,-ny:ny,-nz) = w*h*h*res(-nx:nx,-ny:ny,-nz)/6 + &
!        Psi(-nx:nx,-ny:ny,-nz)
!   Psi(-nx:nx,-ny:ny,nz) = w*h*h*res(-nx:nx,-ny:ny,nz)/6 + &
!        Psi(-nx:nx,-ny:ny,nz)
end select

!The following uses extra (wasted) computing, but does not use forall
!Psi(-nx+1:nx-1,-ny+1:ny-1,-nz+1:nz-1) = &
!  w*h*h*res(-nx+1:nx-1,-ny+1:ny-1,-nz+1:nz-1)/6 + &
!  Psi(-nx+1:nx-1,-ny+1:ny-1,-nz+1:nz-1)

end subroutine relax
!------------------------------------------------------------------------------
subroutine l2n(norm)
use data
implicit none
real*8 :: norm

!norm=sqrt(sum(m*m))
norm=sum(abs(m))/((2*nx+1)*(2*ny+1)*(2*nz+1))
!norm = (2*nx+1)*(2*ny+1)*(2*nz+1)

end subroutine l2n
!------------------------------------------------------------------------------
subroutine l2n_bdry(norm)
use data
implicit none
real*8 :: norm
integer :: points

norm=0
points = 0
do i=-nx, nx
   do j=-ny, ny
      do k=-nz, nz
         if (.not. (abs(i)<nx .and. abs(j)<ny .and. abs(k)<nz))  then
              norm = norm + abs(m(i,j,k))
              points = points + 1
			end if
      end do
   end do
end do
write(6,*) points,norm
norm = norm / (8*(nx*ny+nx*nz+ny*nz)+2)

end subroutine l2n_bdry
!------------------------------------------------------------------------------
subroutine dump(ll,mm,name)
use data
implicit none
integer ll,mm,o
character*(*) name

open(40,file=name)
do o = -nz, nz
   write(40,'(1P,2e14.4)') o*h,m(ll,mm,o)
end do

close(40)


end subroutine dump
